"""Service app and startup function."""
import os
import warnings
from os import path
import logging
import yaml
from typing import List
from fastapi import FastAPI
from ml_base.utilities import ModelManager

from rest_model_service import __version__
from rest_model_service.configuration import Configuration, Model
from rest_model_service.helpers import load_type
from rest_model_service.routes import get_root, get_models, PredictionController  # noqa: F401,E402
from rest_model_service.schemas import Error, ModelMetadataCollection

logger = logging.getLogger(__name__)


def create_app(service_title: str,  models: List[Model]) -> FastAPI:
    """Create instance of FastAPI app and return it."""
    app: FastAPI = FastAPI(title=service_title,
                           version=__version__)

    # adding the routes like this to avoid a circular dependency
    app.add_api_route("/", get_root, methods=["GET"])

    app.add_api_route("/api/models", get_models, methods=["GET"],
                      response_model=ModelMetadataCollection,
                      responses={
                          500: {"model": Error}
                      })

    # loading the models into the ModelManager singleton instance
    model_manager = ModelManager()

    for model in models:
        # loading the model's class
        model_class = load_type(model.class_path)

        # instantiating the model object from the class
        model_instance = model_class()

        # adding the model instance to the ModelManager
        model_manager.add_model(model_instance)

        decorators = model.decorators if model.decorators is not None else []

        # initializing decorators for the model
        for decorator in decorators:
            # loading the decorator's class
            decorator_class = load_type(decorator.class_path)

            # instantiating the decorator object from the class
            if decorator.configuration is not None:
                decorator_instance = decorator_class(**decorator.configuration)
            else:
                decorator_instance = decorator_class()

            # adding the decorator to the model in the ModelManager
            model_manager.add_decorator(model.qualified_name, decorator_instance)

        # creating an endpoint for each model, if the configuration allows it
        if model.create_endpoint:
            model = model_manager.get_model(model.qualified_name)

            controller = PredictionController(model=model)
            controller.__call__.__annotations__["data"] = model.input_schema

            app.add_api_route("/api/models/{}/prediction".format(model.qualified_name),
                              controller,
                              methods=["POST"],
                              response_model=model.output_schema,
                              description=model.description,
                              responses={
                                  400: {"model": Error},
                                  500: {"model": Error}
                              })
        else:
            logger.info("Skipped creating an endpoint for model: {}".format(model.qualified_name))

    return app


# getting the path to the configuration file
# defaults to "rest_config.yaml" in the current directory if not provided
if os.environ.get("REST_CONFIG") is not None:
    file_path = os.environ["REST_CONFIG"]
else:
    file_path = "rest_config.yaml"

# loading the configuration
if path.exists(file_path) and path.isfile(file_path):
    with open(file_path) as file:
        configuration = yaml.full_load(file)
    configuration = Configuration(**configuration)
    app = create_app(configuration.service_title, configuration.models)
else:
    # if there is no configuration file or it is not found, then create an empty app and raise a warning
    app = create_app("REST Model Service", [])
    warnings.warn("Could not find configuration file '{}', service has no models loaded.".format(file_path))



"""Script that generates an OpenAPI schema."""
import os
import sys
import traceback
import argparse
import yaml

from rest_model_service.main import app


def main() -> None:
    """Entry point for the cli tool."""
    argument_parser = argparse.ArgumentParser(description="Generate an OpenAPI schema.")
    argument_parser.add_argument("--output_file", type=str, help="Path of output file.")

    args = argument_parser.parse_args()

    try:
        with open(args.output_file, "w") as file:
            yaml.dump(app.openapi(), file)
    except Exception:
        traceback.print_exc()
        sys.exit(os.EX_SOFTWARE)

    sys.exit(os.EX_OK)


if __name__ == "__main__":
    main()



"""Routes for the service."""
from fastapi.responses import JSONResponse
from starlette.responses import RedirectResponse

from ml_base import MLModel
from ml_base.ml_model import MLModelSchemaValidationException
from ml_base.utilities import ModelManager

from rest_model_service.schemas import ModelMetadataCollection, Error


async def get_root():   # noqa: ANN201
    """Root of API."""
    response = RedirectResponse(url='/docs')
    return response


async def get_models():   # noqa: ANN201
    """List of models available."""
    try:
        # instantiating ModelManager singleton
        model_manager = ModelManager()

        # retrieving the model metadata from the model manager
        model_metadata_collection = model_manager.get_models()
        model_metadata_collection = ModelMetadataCollection(**{"models": model_metadata_collection}).dict()
        return JSONResponse(status_code=200, content=model_metadata_collection)
    except Exception as e:
        error = Error(type="ServiceError", message=str(e)).dict()
        return JSONResponse(status_code=500, content=error)


class PredictionController(object):
    """Callable class that hosts an instance of a model.

    .. note::
       This class is designed to be used as a route in the FastAPI application.

    """

    def __init__(self, model: MLModel) -> None:  # noqa: ANN101
        """Initialize the controller."""
        self._model = model

    def __call__(self, data):  # noqa: ANN001,ANN204,ANN101
        """Make a prediction with a model."""
        try:
            prediction = self._model.predict(data).dict()
            return JSONResponse(status_code=200, content=prediction)
        except MLModelSchemaValidationException as e:
            error = Error(type="SchemaValidationError", message=str(e)).dict()
            return JSONResponse(status_code=400, content=error)
        except Exception as e:
            error = Error(type="ServiceError", message=str(e)).dict()
            return JSONResponse(status_code=500, content=error)



"""Configuration for the service."""
from typing import List, Dict, Any, Optional
from pydantic import BaseModel, Field


class ModelDecorator(BaseModel):
    """Settings for a decorator for a model instance in the service."""

    class_path: str = Field(description="Class path of the decorator class.")
    configuration: Optional[Dict[Any, Any]] = Field(description="Configuration to initialize decorator instance.")


class Model(BaseModel):
    """Settings for a single model in the service."""

    qualified_name: str = Field(description="Qualified name of the model.")
    class_path: str = Field(description="Class path of the model class.")
    create_endpoint: bool = Field(description="Whether or not to create an endpoint for the model.")
    decorators: Optional[List[ModelDecorator]] = Field(description="List of decorators to attach to model.")


class Configuration(BaseModel):
    """Configuration for the service."""

    service_title: str = "RESTful Model Service"
    models: List[Model]



"""Schemas used by the service."""
from typing import List
from pydantic import BaseModel, Field


class ModelMetadata(BaseModel):
    """Metadata of a model."""

    display_name: str = Field(description="The display name of the model.")
    qualified_name: str = Field(description="The qualified name of the model.")
    description: str = Field(description="The description of the model.")
    version: str = Field(description="The version of the model.")


class ModelMetadataCollection(BaseModel):
    """Collection of model metadata."""

    models: List[ModelMetadata] = Field(description="A collection of model description.")


class Error(BaseModel):
    """Error."""

    type: str = Field(description="The type of error.")
    message: str = Field(description="The error message.")



from os import path
from io import open
from setuptools import setup, find_packages


def load_file(file_name):
    here = path.abspath(path.dirname(__file__))
    with open(path.join(here, file_name)) as f:
        return f.read()


setup(name="rest_model_service",
      version=load_file("rest_model_service/version.txt"),
      author="Brian Schmidt",
      author_email="6666331+schmidtbri@users.noreply.github.com",
      description="RESTful service for hosting machine learning models.",
      long_description=load_file("README.md"),
      long_description_content_type="text/markdown",
      url="https://github.com/schmidtbri/rest-model-service",
      license="BSD",
      packages=find_packages(exclude=["tests", "*tests", "tests*"]),
      entry_points={
          'console_scripts': [
              'generate_openapi=rest_model_service.generate_openapi:main',
          ]
      },
      python_requires=">=3.5",
      install_requires=["ml-base>=0.2.0", "fastapi", "uvicorn", "pyyaml"],
      tests_require=['pytest', 'pytest-html', 'pylama', 'coverage', 'coverage-badge', 'bandit', 'safety', "pytype",
                     "flake8-annotations"],
      package_data={
          "rest_model_service": [
                  "version.txt"
            ]
      },
      classifiers=[
          "Programming Language :: Python :: 3",
          "Intended Audience :: Developers",
          "Operating System :: OS Independent"
      ],
      project_urls={
          "Documentation": "https://schmidtbri.github.io/rest-model-service/",
          "Source Code": "https://github.com/schmidtbri/rest-model-service",
          "Tracker": "https://github.com/schmidtbri/rest-model-service/issues"
      },
      keywords=[
          "machine learning", "REST", "service"
      ])



"""RESTful service for hosting machine learning models."""

from os.path import abspath, dirname, join
from ml_base.ml_model import MLModel
from ml_base.decorator import MLModelDecorator
from ml_base.utilities.model_manager import ModelManager

__all__ = ["MLModel", "MLModelDecorator", "ModelManager"]


try:
    print(join(abspath(dirname(__file__)), "version.txt"))
    with open(join(abspath(dirname(__file__)), "version.txt"), encoding="utf-8") as f:
        __version__ = f.read()
except Exception:
    __version__ = "N/A"



"""Helper functions."""
from typing import Type
import importlib


def load_type(class_path: str) -> Type:
    """Load a type.

    Args:
        class_path: Path of the type.

    Returns:
        Type

    """
    # splitting the class_path into module path and class name
    module_path = ".".join(class_path.split(".")[:-1])
    class_name = class_path.split(".")[-1]

    # importing the model class
    _module = importlib.import_module(module_path)
    _class = getattr(_module, class_name)
    return _class
